/*
 * i2c-algo-xilinx.c
 *
 * Xilinx IIC Adapter component to interface IIC component to Linux
 *
 * Author: MontaVista Software, Inc.
 *         source@mvista.com
 *
 * 2002 (c) MontaVista, Software, Inc.  This file is licensed under the terms
 * of the GNU General Public License version 2.  This program is licensed
 * "as is" without any warranty of any kind, whether express or implied.
 */

/*
 * I2C drivers are split into two pieces: the adapter and the algorithm.
 * The adapter is responsible for actually manipulating the hardware and
 * the algorithm is the layer above that that handles the higher level
 * tasks such as transmitting or receiving a buffer.  The best example
 * (in my opinion) of this is the bit banging algorithm has a number of
 * different adapters that can plug in under it to actually wiggle the
 * SDA and SCL.
 *
 * The interesting part is that the drivers Xilinx provides with their
 * IP are also split into two pieces where one part is the OS
 * independent code and the other part is the OS dependent code.  All of
 * the other sources in this directory are the OS independent files as
 * provided by Xilinx with no changes made to them.
 *
 * As it turns out, this maps quite well into the I2C driver philosophy.
 * This file is the I2C algorithm that communicates with the Xilinx OS
 * independent function that will serve as our I2C adapter.  The
 * unfortunate part is that the term "adapter" is overloaded in our
 * context.  Xilinx refers to the OS dependent part of a driver as an
 * adapter.  So from an I2C driver perspective, this file is not an
 * adapter; that role is filled by the Xilinx OS independent files.
 * From a Xilinx perspective, this file is an adapter; it adapts their
 * OS independent code to Linux.
 *
 * Another thing to consider is that the Xilinx OS dependent code knows
 * nothing about Linux I2C adapters, so even though this file is billed
 * as the I2C algorithm, it takes care of the i2c_adapter structure.
 *
 * Fortunately, naming conventions will give you a clue as to what comes
 * from where.  Functions beginning with XIic_ are provided by the
 * Xilinx OS independent files.  Functions beginning with i2c_ are
 * provided by the I2C Linux core.  All functions in this file that are
 * called by Linux have names that begin with xiic_.  The functions in
 * this file that have Handler in their name are registered as callbacks
 * with the underlying Xilinx OS independent layer.  Any other functions
 * are static helper functions.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/ioctl.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/i2c.h>
#include <xbasic_types.h>
#include <linux/delay.h>
#include "xiic.h"
#include "xiic_i.h"
#include "xiic_fpga.h"

#include "xparameters.h"
#include "xio.h"
#include <gpio_struct.h>
#include <gpioiocts.h>
#include <cmc_gpio.h>


#define DCRN_UIC0_BASE	0x0C0
#define DCRN_UIC0_SR    (DCRN_UIC0_BASE + 0x0)
#define DCRN_UIC0_ER    (DCRN_UIC0_BASE + 0x2)
#define DCRN_UIC0_CR    (DCRN_UIC0_BASE + 0x3)
#define DCRN_UIC0_PR    (DCRN_UIC0_BASE + 0x4)
#define DCRN_UIC0_TR    (DCRN_UIC0_BASE + 0x5)
#define DCRN_UIC0_MSR   (DCRN_UIC0_BASE + 0x6)
#define DCRN_UIC0_VR    (DCRN_UIC0_BASE + 0x7)
#define DCRN_UIC0_VCR   (DCRN_UIC0_BASE + 0x8)


MODULE_AUTHOR("MontaVista Software, Inc. <source@mvista.com>");
MODULE_DESCRIPTION("Xilinx IIC driver");
MODULE_LICENSE("GPL");
//2.6.24
//MODULE_PARM(scan, "i");
MODULE_PARM_DESC(scan, "Scan for active chips on the bus");
static int scan = 0;		/* have a look at what's hanging 'round */

/* SAATODO: actually use these? */
#define XIIC_TIMEOUT	100
#define XIIC_RETRY		3
#define	MAX_XIIC_RETRY	16
#define SMBUS_TIMEOUT	15

// Uncomment the following line if you want to mask SMBUS timeout interrupts.
//#define MASK_CMC_IOM_I2C_SMBUS_INTERRUPTS

/* Our private per device data. */
struct xiic_dev {
	struct i2c_adapter adap;	/* The Linux I2C core data  */
	struct xiic_dev *next_dev;	/* The next device in dev_list */
	struct completion complete;	/* for waiting for interrupts */
	int index;		/* Which interface is this */
	u32 save_BaseAddress;	/* Saved physical base address */
	u32 BaseAddress;	/* Saved virtual base address */
	u32 FPGABaseAddress;	/* Saved virtual FPGA base address */
	unsigned int irq;	/* device IRQ number    */
	/*
	 * The underlying OS independent code needs space as well.  A
	 * pointer to the following XIic structure will be passed to
	 * any XIic_ function that requires it.  However, we treat the
	 * data as an opaque object in this file (meaning that we never
	 * reference any of the fields inside of the structure).
	 */
	XIic Iic;
	XStatus interrupt_status;
	/*
	 * The following bit fields are used to keep track of what
	 * all has been done to initialize the xiic_dev to make
	 * error handling out of probe() easier.
	 */
	unsigned int reqirq:1;	/* Has request_irq() been called? */
	unsigned int remapped:1;	/* Has ioremap() been called? */
	unsigned int started:1;	/* Has XIic_Start() been called? */
	unsigned int added:1;	/* Has i2c_add_adapter() been called? */

  unsigned long itime;
};

/* List of devices we're handling and a lock to give us atomic access. */
static struct xiic_dev *dev_list = NULL;
static spinlock_t dev_lock = SPIN_LOCK_UNLOCKED;


/* SAATODO: This function will be moved into the Xilinx code. */
/*****************************************************************************/
/**
*
* Lookup the device configuration based on the iic instance.  The table
* XIic_ConfigTable contains the configuration info for each device in the system.
*
* @param Instance is the index of the interface being looked up.
*
* @return
*
* A pointer to the configuration table entry corresponding to the given
* device ID, or NULL if no match is found.
*
* @note
*
* None.
*
******************************************************************************/
XIic_Config *XIic_GetConfig(int Instance)
{
	if (Instance < 0 || Instance >= XPAR_XIIC_NUM_INSTANCES)
	{
		return NULL;
	}

	return &XIic_ConfigTable[Instance];
}

void XIic_ResetI2cCore(XIic *IicPtr, unsigned int index)
{
	// Issue the I2C core reset
	XIo_Out16(IicPtr->BaseAddress + CMC_GPIO_CORE_RESET  , 0x000a);
	// Re-initialize the I2c registers
	if (XIic_Start(IicPtr) != XST_SUCCESS) {
		printk(KERN_ERR "xiic_xfer, could not restart the device, index: %d.\n", index);
	}
}

static int
xiic_xfer(struct i2c_adapter *i2c_adap, struct i2c_msg msgs[], int num)
{
	struct xiic_dev *dev = (struct xiic_dev *) i2c_adap;
	struct i2c_msg *pmsg;
	u32 options;
	int i, retries;
	XStatus Status;
	static unsigned long mj;
	unsigned long now;
	int old_nice;
	int ret=0;
	u16 ControlReg;

	old_nice = task_nice(current);
#ifndef AVOID_XIIC_SMBUS_TIMEOUT
	/*  Somekind of Workaround for SMBUS Timeout May not require with new driver changes*/
	set_user_nice(current, -19);
#endif
	/* Make sure the RecvBufer is NULL beofre any operation */
	/* TxError handling is based on RecvBuffer*/
	dev->Iic.RecvBufferPtr = NULL;



#ifdef AVOID_XIIC_SMBUS_TIMEOUT
	if (num>2 || num <= 0){
		printk(KERN_WARNING "Invalid Number of Messages in I2C operation\n");
		printk(KERN_WARNING "Messages in I2C operation should be 1 or 2\n");
		ret = -EINVAL; 
		goto FUNC_END;
		}
	if ((2 == num) && (!(msgs[0].flags & I2C_M_RD)) && (msgs[1].flags & I2C_M_RD)){
		pmsg      = &msgs[1];
		dev->Iic.RecvByteCount = pmsg->len ;
		dev->Iic.RecvBufferPtrRepeated = pmsg->buf ;
		}

			pmsg = &msgs[0];
			
			if (!pmsg->len) /* If length is zero */
			{		
				return num;
			}  /* on to the next request. */
#else
	for (i = 0; i < num; i++) {
		pmsg = &msgs[i];

		if (!pmsg->len)	/* If length is zero */
		     continue;	/* on to the next request. */
#endif

		ControlReg = XIic_CheckI2CBusState(&dev->Iic);
		if (ControlReg){
			printk(KERN_WARNING
						"%s #%d: Xilinx I2C Bus NOT IDLE (SDA:0x%x SCL:%x)\n",
						dev->adap.name, dev->index, (ControlReg & 0x0008)>>3,(ControlReg & 0x0002)>>1);
			ret = -EBUSY;				
			goto FUNC_END;
			}


		options = 0;
		if (pmsg->flags & I2C_M_TEN)
			options |= XII_SEND_10_BIT_OPTION;
#ifdef AVOID_XIIC_SMBUS_TIMEOUT		
	       	if (2 == num)
			options |= XII_REPEATED_START_OPTION;
#else
		if (i != num-1)
			options |= XII_REPEATED_START_OPTION;
#endif
		XIic_SetOptions(&dev->Iic, options);

		now = jiffies;
		if (XIic_SetAddress(&dev->Iic, XII_ADDR_TO_SEND_TYPE,
				    pmsg->addr) != XST_SUCCESS) {
			printk(KERN_WARNING
			       "%s #%d: Could not set address to 0x%2x.\n",
			       dev->adap.name, dev->index, pmsg->addr);
			ret = -EXDEV; 
			goto FUNC_END;
		}
		dev->interrupt_status = ~(XStatus) 0;
		/*
		 * The Xilinx layer does not handle bus busy conditions yet
		 * so this code retries a request up to 16 times if it
		 * receives a bus busy condition.  If and when the underlying
		 * code is enhanced, the retry code can be removed.
		 */
		retries = MAX_XIIC_RETRY;
		if (pmsg->flags & I2C_M_RD) {
			while (retries-- > 0) {
				spin_lock_irq(&dev->Iic.lock);
				Status = XIic_MasterRecv(&dev->Iic, pmsg->buf, pmsg->len);
				spin_unlock_irq(&dev->Iic.lock);
				if (Status == XST_IIC_BUS_BUSY) {
					set_current_state(TASK_INTERRUPTIBLE);
					schedule_timeout(HZ/10);
				} else break;
			}
		} else {
			while (retries-- > 0) {
				spin_lock_irq(&dev->Iic.lock);
				Status = XIic_MasterSend(&dev->Iic, pmsg->buf, pmsg->len);
				spin_unlock_irq(&dev->Iic.lock);
				if (Status == XST_IIC_BUS_BUSY) {
					set_current_state(TASK_INTERRUPTIBLE);
					schedule_timeout(HZ/10);
				} else break;
			}
		}
		if (Status != XST_SUCCESS) {
			if (Status == XST_IIC_BUS_BUSY)
			{
				// I2C bus is busy, we log the message
				// into /var/log/messages here by using printk.
				printk(KERN_WARNING
			       		"%s #%d: I2C Bus Busy error %d.\n",
			       		dev->adap.name, dev->index, Status);
				printk(KERN_WARNING "I2C core reset issued for device index: %d.\n", dev->index);
				// The bus is dead here, in order to recover
				// the bus, we need to issue
				// the I2C bus core reset and re-initialize
				// the I2C registers. 
				XIic_ResetI2cCore(&dev->Iic, dev->index);
				ret = -EBUSY;				
			} else
			{
				printk(KERN_WARNING
			       		"%s #%d: Unexpected error %d.\n",
			       		dev->adap.name, dev->index, Status);
				ret = -ENODATA;
			}
			goto FUNC_END;
		}

		// This is only the temporary workaround for the
		// SMBUS timeout problem.
		// This should be removed for the final release
#if 0
		for (k = 0; k < SMBUS_TIMEOUT; k++) {
			if (dev->interrupt_status == XST_SUCCESS)
				break;
			else
				mdelay(2);
		}
		if (dev->interrupt_status != XST_SUCCESS) {
			printk(KERN_WARNING
		       	"%s #%d: Could not talk to device 0x%2x (%d).\n",
		       	dev->adap.name, dev->index, pmsg->addr,
		       	dev->interrupt_status);
			goto badio;
		}
#endif
		// This should be the final codes for the final release
		if (!wait_for_completion_timeout(&dev->complete, 1000))
		{
			if (dev->interrupt_status != XST_SUCCESS) {
				printk(KERN_WARNING
			       	"%s #%d: Could not talk to device 0x%2x (%d).\n",
			       	dev->adap.name, dev->index, pmsg->addr,
			       	dev->interrupt_status);
				ret = -ETIME;
				goto FUNC_END;
			}
		}
		if ((jiffies - now) > mj) {
		  mj = jiffies - now;
		  //printk(KERN_INFO"xiic: max jiffies: %lu %lu\n", mj, jiffies - dev->itime);
		}
		if (dev->interrupt_status  ==  XII_SLAVE_NO_ACK_EVENT) {
#if 0	//put this back in when we fix all the isr/i2c stuff
				printk(KERN_WARNING
			       	"%s #%d: Slave NACK 0x%2x (%d).\n",
			       	dev->adap.name, dev->index, pmsg->addr,
			       	dev->interrupt_status);
#endif
			ret = -ENODEV;
			goto FUNC_END;
		}
#ifndef AVOID_XIIC_SMBUS_TIMEOUT
	}
#endif
#ifndef AVOID_XIIC_SMBUS_TIMEOUT
	/*  Somekind of Workaround for SMBUS Timeout May not require with new driver changes*/	
	set_user_nice(current, old_nice);
#endif
	return num;
 FUNC_END:
#ifndef AVOID_XIIC_SMBUS_TIMEOUT
	 /*  Somekind of Workaround for SMBUS Timeout May not require with new driver changes*/  
	set_user_nice(current, old_nice);
#endif
	return ret;
}

static int
xiic_algo_control(struct i2c_adapter *adapter,
		  unsigned int cmd, unsigned long arg)
{
    int index;
    int i;
    XIic_Config *cfg;
    struct xiic_dev *dev;

    switch(cmd)
    {
	case IOCTL_INIT_XILINX_IOM1:
	{
		index = 0;
		cfg = XIic_GetConfig(index);
		dev = dev_list;
		for ( i = 0; i < 5; i++)
		{
			if (dev->index == index)
			{
				break;
			}
			dev = dev->next_dev;
		}
		/* Tell the Xilinx code to re-start this IIC bus. */
		spin_lock_irq(&dev->Iic.lock);
		if (XIic_Start(&dev->Iic) != XST_SUCCESS) {
			printk(KERN_ERR "%s #%d: Could not start device.\n",
		       	dev->adap.name, dev->index);
			spin_unlock_irq(&dev->Iic.lock);
			return -ENODEV;
		}
		spin_unlock_irq(&dev->Iic.lock);
		break;
	}
	case IOCTL_INIT_XILINX_IOM2:
	{
		index = 1;
		cfg = XIic_GetConfig(index);
		dev = dev_list;
		for ( i = 0; i < 5; i++)
		{
			if (dev->index == index)
			{
				break;
			}
			dev = dev->next_dev;
		}
		/* Tell the Xilinx code to re-start this IIC bus. */
		spin_lock_irq(&dev->Iic.lock);
		if (XIic_Start(&dev->Iic) != XST_SUCCESS) {
			printk(KERN_ERR "%s #%d: Could not start device.\n",
		       	dev->adap.name, dev->index);
			spin_unlock_irq(&dev->Iic.lock);
			return -ENODEV;
		}
		spin_unlock_irq(&dev->Iic.lock);
		break;
	}
	case IOCTL_INIT_XILINX_PSU1:
	{
		index = 2;
		cfg = XIic_GetConfig(index);
		dev = dev_list;
		for ( i = 0; i < 5; i++)
		{
			if (dev->index == index)
			{
				break;
			}
			dev = dev->next_dev;
		}
		/* Tell the Xilinx code to re-start this IIC bus. */
		spin_lock_irq(&dev->Iic.lock);
		if (XIic_Start(&dev->Iic) != XST_SUCCESS) {
			printk(KERN_ERR "%s #%d: Could not start device.\n",
		       	dev->adap.name, dev->index);
			spin_unlock_irq(&dev->Iic.lock);
			return -ENODEV;
		}
		spin_unlock_irq(&dev->Iic.lock);
		break;
	}
	case IOCTL_INIT_XILINX_PSU2:
	{
		index = 3;
		cfg = XIic_GetConfig(index);
		dev = dev_list;
		for ( i = 0; i < 5; i++)
		{
			if (dev->index == index)
			{
				break;
			}
			dev = dev->next_dev;
		}
		/* Tell the Xilinx code to re-start this IIC bus. */
		spin_lock_irq(&dev->Iic.lock);
		if (XIic_Start(&dev->Iic) != XST_SUCCESS) {
			printk(KERN_ERR "%s #%d: Could not start device.\n",
		       	dev->adap.name, dev->index);
			spin_unlock_irq(&dev->Iic.lock);
			return -ENODEV;
		}
		spin_unlock_irq(&dev->Iic.lock);
		break;
	}
	case IOCTL_INIT_XILINX_KVM:
	{
		index = 4;
		cfg = XIic_GetConfig(index);
		dev = dev_list;
		for ( i = 0; i < 5; i++)
		{
			if (dev->index == index)
			{
				break;
			}
			dev = dev->next_dev;
		}
		/* Tell the Xilinx code to re-start this IIC bus. */
		spin_lock_irq(&dev->Iic.lock);
		if (XIic_Start(&dev->Iic) != XST_SUCCESS) {
			printk(KERN_ERR "%s #%d: Could not start device.\n",
		       	dev->adap.name, dev->index);
			spin_unlock_irq(&dev->Iic.lock);
			return -ENODEV;
		}
		spin_unlock_irq(&dev->Iic.lock);
		break;
	}
	default:
		break;
     } // end of switch
	return 0;
}

static u32
xiic_bit_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_SMBUS_EMUL | I2C_FUNC_10BIT_ADDR |
	    I2C_FUNC_PROTOCOL_MANGLING;
}

static struct i2c_algorithm xiic_algo = {
        .master_xfer    = xiic_xfer,
        .functionality  = xiic_bit_func,
	// 2.6.24, chl chl, we keep this member in
	// the structure and put the member back in linux/include/i2c.h
	.algo_control   = xiic_algo_control,
		.smbus_xfer     = NULL
};

/*
 * This routine is registered with the OS as the function to call when
 * the IIC interrupts.  It in turn, calls the Xilinx OS independent
 * interrupt function.  The Xilinx OS independent interrupt function
 * will in turn call any callbacks that we have registered for various
 * conditions.
 */
irqreturn_t
xiic_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{

	irqreturn_t status;

	struct xiic_dev *dev = dev_id;

	// DF294497
	//spin_lock_irq(&dev->Iic.lock);
	// Check for spurious interrupt before we process the
	// interrupts.
	if (dev == NULL)
        {
	    printk("Spurious interrupt, NULL pointer.\n"); 
	    return IRQ_HANDLED;
 	}
	if ( (dev->save_BaseAddress != XPAR_OPB_IIC_0_BASEADDR) &&
	 (dev->save_BaseAddress != XPAR_OPB_IIC_1_BASEADDR) &&
	 (dev->save_BaseAddress != XPAR_OPB_IIC_2_BASEADDR) &&
	 (dev->save_BaseAddress != XPAR_OPB_IIC_3_BASEADDR) &&
	 (dev->save_BaseAddress != XPAR_OPB_IIC_4_BASEADDR) )
	 {
	   // Return IRQ_NONE since it might be someone inserted/removed
	   // the device
           return IRQ_NONE;
	 }
	
	dev->itime = jiffies;
	//printk("Physical address: 0x%x\n", dev->save_BaseAddress);
	status = XIic_InterruptHandler(&dev->Iic, irq);
	//spin_unlock_irq(&dev->Iic.lock);

	return status;
}

static void
RecvHandler(void *CallbackRef, int ByteCount)
{
	struct xiic_dev *dev = (struct xiic_dev *) CallbackRef;

	if (ByteCount == 0) {
		dev->interrupt_status = XST_SUCCESS;
		complete(&dev->complete);
	}
}
static void
SendHandler(void *CallbackRef, int ByteCount)
{
	struct xiic_dev *dev = (struct xiic_dev *) CallbackRef;

	if (ByteCount == 0) {
		dev->interrupt_status = XST_SUCCESS;
		complete(&dev->complete);
	}
}
static void
StatusHandler(void *CallbackRef, XStatus Status)
{
	struct xiic_dev *dev = (struct xiic_dev *) CallbackRef;

	dev->interrupt_status = Status;
	complete(&dev->complete);
}

#if 0
static void
xiic_inc_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	// chl chl, not needed for 2.6 kernel
	//MOD_INC_USE_COUNT;
#endif
}
static void
xiic_dec_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	// chl chl, not needed for 2.6 kernel
	//MOD_DEC_USE_COUNT;
#endif
}
#endif

static void
remove_head_dev(void)
{
	struct xiic_dev *dev;
	XIic_Config *cfg;

	/* Pull the head off of dev_list. */
	spin_lock(&dev_lock);
	dev = dev_list;
	dev_list = dev->next_dev;
	spin_unlock(&dev_lock);

	/*
	 * If we've told the core I2C code about this dev, tell
	 * the core I2C code to forget the dev.
	 */
	if (dev->added) {
		/*
		 * If an error is returned, there's not a whole lot we can
		 * do.  An error has already been printed out so we'll
		 * just keep trundling along.
		 */
		(void) i2c_del_adapter(&dev->adap);
	}

	/* Tell the Xilinx code to take this IIC interface down. */
	if (dev->started) {
		while (XIic_Stop(&dev->Iic) != XST_SUCCESS) {
			/* The bus was busy.  Retry. */
			printk(KERN_WARNING
			       "%s #%d: Could not stop device.  Will retry.\n",
			       dev->adap.name, dev->index);
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ / 2);
		}
	}

	/*
	 * Now that the Xilinx code isn't using the IRQ or registers,
	 * unmap the registers and free the IRQ.
	 */
	if (dev->remapped) {
		cfg = XIic_GetConfig(dev->index);
		iounmap((void *) cfg->BaseAddress);
		iounmap((void *) cfg->FPGABaseAddress);
		cfg->BaseAddress = dev->save_BaseAddress;
	};
	if (dev->reqirq) {
		free_irq(dev->irq, dev);
	}

	kfree(dev);
}

static int __init
probe(int index)
{
	static const unsigned long remap_size
	    = XPAR_IIC_0_HIGHADDR - XPAR_IIC_0_BASEADDR + 1;
	struct xiic_dev *dev;
	XIic_Config *cfg;
	unsigned int irq;
	int retval;


	switch (index) {
#if defined(XPAR_INTC_0_IIC_0_VEC_ID)
	case 0:
		// IOM123
		irq = 65;
		// Enable external IRQ 6
		//mtdcr(DCRN_UIC2_ER, mfdcr(DCRN_UIC2_ER) | 0x40000000 );
		//mtdcr(0xe2, mfdcr(0xe2) | 0x40000000 );
		break;
#endif				/* 0 */
#if defined(XPAR_INTC_0_IIC_1_VEC_ID)
	case 1:
		// IOM 456
		irq = 65;
		// Enable external IRQ 6
		//mtdcr(0xe2, mfdcr(0xe2) | 0x40000000 );
		break;
#endif				/* 1 */
#if defined(XPAR_INTC_0_IIC_2_VEC_ID)
	case 2:
		// PSU 123
		irq = 50;
		// Enable external IRQ 7
		//mtdcr(DCRN_UIC1_ER, mfdcr(DCRN_UIC1_ER) | 0x00002000 );
		//mtdcr(0xd2, mfdcr(0xd2) | 0x00002000 );
		break;
#endif				/* 2 */
#if defined(XPAR_INTC_0_IIC_3_VEC_ID)
	case 3:
		// PSU 456
		irq = 50;
		// Enable external IRQ 7
		//mtdcr(DCRN_UIC1_ER, mfdcr(DCRN_UIC1_ER) | 0x00002000 );
		//mtdcr(0xd2, mfdcr(0xd2) | 0x00002000 );
		break;
#endif				/* 3 */
#if defined(XPAR_INTC_0_IIC_4_VEC_ID)
	case 4:
		// KVM
		irq = 27;
		break;
#endif				/* 4 */
	default:
		return -ENODEV;
	}

	/* Find the config for our device. */
	
	cfg = XIic_GetConfig(index);
	
	if (!cfg)
		return -ENODEV;

	/* Allocate the dev and zero it out. */
	dev = (struct xiic_dev *) kmalloc(sizeof (struct xiic_dev), GFP_KERNEL);
	if (!dev) {
		printk(KERN_ERR "%s #%d: Could not allocate device.\n",
		       "Xilinx IIC adapter", index);
		return -ENOMEM;
	}
	memset(dev, 0, sizeof (struct xiic_dev));
	spin_lock_init(&dev->Iic.lock);
	strcpy(dev->adap.name, "Xilinx IIC adapter");
	dev->index = index;

	/* Make it the head of dev_list. */
	spin_lock(&dev_lock);
	dev->next_dev = dev_list;
	dev_list = dev;
	spin_unlock(&dev_lock);

	init_completion(&dev->complete);

	/* Change the addresses to be virtual; save the old ones to restore. */
	dev->save_BaseAddress = cfg->BaseAddress;
	cfg->BaseAddress = (u32) ioremap(dev->save_BaseAddress, remap_size);
	cfg->FPGABaseAddress = (u32) ioremap(XPAR_MY_OPB_GPIO_0_BASEADDR_0, 
		XPAR_MY_OPB_GPIO_0_HIGHADDR_0 - XPAR_MY_OPB_GPIO_0_BASEADDR_0);
	cfg->dev = index;
	dev->BaseAddress = cfg->BaseAddress;
	dev->FPGABaseAddress = cfg->FPGABaseAddress;
	dev->remapped = 1;


	/* Tell the Xilinx code to bring this IIC interface up. */
	if (XIic_Initialize(&dev->Iic, cfg->DeviceId) != XST_SUCCESS) {
		printk(KERN_ERR "%s #%d: Could not initialize device.\n",
		       dev->adap.name, dev->index);
		remove_head_dev();
		return -ENODEV;
	}
	XIic_SetRecvHandler(&dev->Iic, (void *) dev, RecvHandler);
	XIic_SetSendHandler(&dev->Iic, (void *) dev, SendHandler);
	XIic_SetStatusHandler(&dev->Iic, (void *) dev, StatusHandler);

#ifdef  MASK_CMC_IOM_I2C_SMBUS_INTERRUPTS
        // BC: Mask the SM timeout interrupts, as we are seeing nobody cared.
        if (65 == irq)
 	{
		u32 fpga_addr;
		u16 val;

		fpga_addr = dev->FPGABaseAddress + IOM_IIER_OFFSET;
		val = (IOM123_SMTO_INTR_EN_MASK|IOM456_SMTO_INTR_EN_MASK);

    		val |= XIo_In16(fpga_addr);
    		XIo_Out16(fpga_addr,val);

	}
#endif

	/* Grab the IRQ */
	// chl chl
	dev->irq = irq;
	//printk(KERN_INFO "Reduced SMBUS T/O jordan driver\n");
		// 2.6.24
		//retval = request_irq(dev->irq, xiic_interrupt, SA_SHIRQ, dev->adap.name, dev);
		retval = request_irq(dev->irq, xiic_interrupt, IRQF_SHARED, dev->adap.name, dev);
		if (retval) {
			printk(KERN_ERR "%s #%d: Could not allocate interrupt %d.\n",
		       		dev->adap.name, dev->index, dev->irq);
			remove_head_dev();
		return retval;
		}

	if (XIic_Start(&dev->Iic) != XST_SUCCESS) {
		printk(KERN_ERR "%s #%d: Could not start device.\n",
		       dev->adap.name, dev->index);
		remove_head_dev();
		return -ENODEV;
	}
	dev->started = 1;

	/* Now tell the core I2C code about our new device. */
	/*
	 * SAATODO: Get a real ID (perhaps I2C_HW_XILINX) after
	 * initial release.  Will need to email lm78@stimpy.netroedge.com
	 * per http://www2.lm-sensors.nu/~lm78/support.html
	 */
	// chl chl, id doesn't exist for 2.6.14 kernel
	// 2.6.24
	//dev->adap.id = I2C_ALGO_EXP | I2C_DRIVERID_EXP0;
	dev->adap.algo = &xiic_algo;
	dev->adap.algo_data = NULL;
	dev->adap.client_register = NULL;
	dev->adap.client_unregister = NULL;
	// chl chl, not needed for 2.6 kernel
	// dev->adap.inc_use = xiic_inc_use;
	// dev->adap.dec_use = xiic_dec_use;
	dev->adap.timeout = XIIC_TIMEOUT;
	dev->adap.retries = XIIC_RETRY;
	dev->reqirq = 1;
	i2c_set_adapdata(&dev->adap, dev);
	retval = i2c_add_adapter(&dev->adap);
	if (retval) {
		printk(KERN_ERR "%s #%d: Could not add I2C adapter.\n",
		       dev->adap.name, dev->index);
		remove_head_dev();
		return retval;
	}
	dev->added = 1;

	printk(KERN_INFO "%s #%d at 0x%08X mapped to 0x%08X, irq=%d\n",
	       dev->adap.name, dev->index,
	       dev->save_BaseAddress, cfg->BaseAddress, dev->irq);

	/* scan bus */
	
	if (scan) {
		int i, anyfound;

		printk(KERN_INFO "%s #%d Bus scan:",
		       dev->adap.name, dev->index);
		/*
		 * The Philips I2C-bus specification defines special addresses:
		 *      0x00       General Call
		 *      0x01       CBUS Address
		 *      0x02       Reserved for different bus format
		 *      0x03       Reserved for future purposes
		 *   0x04 to 0x07  Hs-mode master code
		 *   0x78 to 0x7B  10-bit slave addressing
		 *   0x7C to 0x7F  Reserved for future purposes
		 *
		 * We don't bother scanning any of these addresses.
		 */
		anyfound = 0;
		for (i = 0x08; i < 0x78; i++) {
			u8 data;
			XStatus Status;
			unsigned long now = jiffies;
			static unsigned long mj;

			if (XIic_SetAddress(&dev->Iic, XII_ADDR_TO_SEND_TYPE,
					    i) != XST_SUCCESS) {
				printk(KERN_WARNING
				       "%s #%d: Could not set address to %2x.\n",
				       dev->adap.name, dev->index, i);
				continue;
			}
			dev->interrupt_status = ~(XStatus) 0;
			Status = XIic_MasterRecv(&dev->Iic, &data, 1);
			if (Status != XST_SUCCESS) {
				printk(KERN_WARNING
				       "%s #%d: Unexpected error %d.\n",
				       dev->adap.name, dev->index, Status);
				continue;
			}
			if (!wait_for_completion_timeout(&dev->complete, 1000))
			{
				if (dev->interrupt_status == XST_SUCCESS) {
					anyfound = 1;
				}
			}
			if ((jiffies - now) >= mj) {
			  mj = jiffies - now;
			  //printk(KERN_INFO "xiic: max jiffies: %lu %lu\n", mj, jiffies - dev->itime);
			}
			if (dev->interrupt_status == XST_SUCCESS) {
				anyfound = 1;
			}
		}
		printk(anyfound ? " responded.\n" : " Nothing responded.\n");
	}

	return 0;
}

#if 0
void xiic_dump_uic()
{
	int temp;
			temp =  mfdcr(0xc0);
			printk("UIC0 status register c0: 0x%x\n", temp);
			temp =  mfdcr(0xd0);
			printk("UIC1 status register d0: 0x%x\n", temp);
			temp =  mfdcr(0xe0);
			printk("UIC2 status register e0: 0x%x\n", temp);
			temp =  mfdcr(0xc6);
			printk("UIC Mask register c6: 0x%x\n", temp);
			temp =  mfdcr(0xd6);
			printk("UIC Mask register d6: 0x%x\n", temp);
			temp =  mfdcr(0xe6);
			printk("UIC Mask register e6: 0x%x\n", temp);
			temp =  mfdcr(0xc2);
			printk("UIC enable register c2: 0x%x\n", temp);
			temp =  mfdcr(0xd2);
			printk("UIC enable register d2: 0x%x\n", temp);
			temp =  mfdcr(0xe2);
			printk("UIC enable register e2: 0x%x\n", temp);
}
#endif
static int __init
xiic_init(void)
{
	int index = 0;

	for (index = 0; index < 5; index++)
		probe(index);

	// while (probe(index++) == 0) ;
	/* If we found at least one, report success. */
	// return (index > 5) ? 0 : -ENODEV;
	return 0;
}

static void __exit
xiic_cleanup(void)
{
	while (dev_list)
		remove_head_dev();
}

int EXPORT_NO_SYMBOLS;

module_init(xiic_init);
module_exit(xiic_cleanup);
